unit lazbridge32;

{$mode objfpc}{$H+}

interface

uses
  {$ifdef LCLGtk2}
  {$define Op32_Supported}
   Cairo, gdk2, gtk2Def, gtk2Proc,
   //Cairo, gdk2, gtkDef, gtkProc, // for versions before rev 26012
  {$endif}
  {$ifdef LCLQt}
  {$define Op32_Supported}
   Qt4, QtObjects, LCLType,
  {$endif}
  {$ifdef LCLWin32}
  {$define Op32_Supported}
   Windows, Win32Extra,
  {$endif}
  {$ifdef LCLCarbon}
  {$define Op32_Supported}
   FPCMacOSAll, CarbonProc, CarbonGDIObjects, CarbonCanvas,
  {$endif}
  Graphics, Classes, SysUtils, opbitmap;


{$i PreMultipliedTable.inc}


procedure OpBitmapPaint32(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer);
Procedure OpBitmapPaint32A(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer;
          SourcePrepared:boolean=false);
Procedure OpBitmapPreMultiply(Source,Target: TOPBitmap);


implementation

{$ifndef Op32_Supported}
uses lazbridge;
{$endif}

Procedure OpBitmapPreMultiply(Source,Target: TOPBitmap);
Var i: integer;
  sarr,tarr: PAPixel32;
Begin
  if (Source.Width=0) or (Source.Height=0) then exit;
  Source.PixelFormat := pf32bit;
  Target.PixelFormat := pf32bit;
  If (Target.Width<>Source.Width) Or (Target.Height<>Source.Height) Then
    Begin;
      Target.Height := 0;
      Target.Width := Source.Width;
      Target.Height := Source.Height;
    End;
  i := 0;
  sarr := TBitmapData32(Source.Data).RawArray;
  tarr := TBitmapData32(Target.Data).RawArray;
  For i:=0 To (Source.Width*Source.Height) Do
    Begin
      tarr^[i].Blue := PreMultipliedTable[sarr^[i].Blue][sarr^[i].Alpha];
      tarr^[i].Red := PreMultipliedTable[sarr^[i].Red][sarr^[i].Alpha];
      tarr^[i].Green := PreMultipliedTable[sarr^[i].Green][sarr^[i].Alpha];
      tarr^[i].Alpha := sarr^[i].Alpha;
    End;
End;


{$ifdef LCLGtk2}
function gdk_cairo_create(drawable: PGdkDrawable): Pcairo_t; cdecl; external gdklib;

Procedure OpBitmapPaint32A(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer;
          SourcePrepared:boolean=false);
Var cr: Pcairo_t;
  surface: Pcairo_surface_t;
  TempBmp: TOPBitmap;
Begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
  If Not SourcePrepared Then
    Begin
      TempBmp := TOPBitmap.Create;
      OpBitmapPreMultiply(Bmp,TempBmp);
      Bmp := TempBmp;
    End Else Bmp.PixelFormat := pf32bit;

  cr := gdk_cairo_create(TGtkDeviceContext(Canvas.Handle).Drawable);
  surface := cairo_image_surface_create_for_data(Bmp.ScanLine[0], CAIRO_FORMAT_ARGB32, Bmp.Width,
             Bmp.Height, Bmp.Data.LineLength);
  If surface <> Nil Then
    Begin
      cairo_set_source_surface(cr, surface, X, Y);
      cairo_paint(cr);
      cairo_surface_destroy(surface);
    End;
  cairo_destroy(cr);
  If Not SourcePrepared Then TempBmp.free;
End;

procedure OpBitmapPaint32( Bmp: TOpBitmap; Canvas:Graphics.TCanvas; X, Y: integer);
var cr: Pcairo_t;
  surface: Pcairo_surface_t;
begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
  Bmp.PixelFormat := pf32bit;
  cr := gdk_cairo_create(TGtkDeviceContext(Canvas.Handle).Drawable);
  surface := cairo_image_surface_create_for_data(Bmp.ScanLine[0], CAIRO_FORMAT_RGB24, Bmp.Width, Bmp.Height, Bmp.Data.LineLength);
  if surface <> nil then
  begin
    cairo_set_source_surface(cr, surface, X, Y);
    cairo_paint(cr);
  end;
  cairo_destroy(cr);
  cairo_surface_destroy(surface);
end;
{$endif}

{$ifdef LCLQt}
procedure OpBitmapPaint32A(Bmp: TOpBitmap;Canvas:Graphics.TCanvas; X, Y: integer;
          SourcePrepared:boolean=false);
var  Image: TQtImage;
 DstQDC: TQtDeviceContext;
 SrcRect, DstRect: TRect;
begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
 Bmp.PixelFormat := pf32bit;
 DstQDC:=TQtDeviceContext(Canvas.Handle);
 DstRect := Bounds(X, Y, Bmp.Width, Bmp.Height);
 SrcRect := Bounds(0,0,Bmp.Width, Bmp.Height);
 if SourcePrepared then
 Image := TQtImage.Create(Bmp.ScanLine[0], Bmp.Width, Bmp.Height, QImageFormat_ARGB32_Premultiplied) else
 Image := TQtImage.Create(Bmp.ScanLine[0], Bmp.Width, Bmp.Height, QImageFormat_ARGB32);
try
 QPainter_drawImage(DstQDC.Widget, PRect(@DstRect), Image.Handle, @SrcRect, QtAutoColor);
 finally
 Image.Free;
 end;
end;

procedure OpBitmapPaint32(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer);
var  Image: TQtImage;
 DstQDC: TQtDeviceContext;
 SrcRect, DstRect: TRect;
begin
 if (Bmp.Width=0) or (Bmp.Height=0) then exit;
 Bmp.PixelFormat := pf32bit;
 DstQDC:=TQtDeviceContext(Canvas.Handle);
 DstRect := Bounds(X, Y, Bmp.Width, Bmp.Height);
 SrcRect := Bounds(0,0,Bmp.Width, Bmp.Height);
Image := TQtImage.Create(Bmp.ScanLine[0], Bmp.Width, Bmp.Height, QImageFormat_RGB32);
try
 QPainter_drawImage(DstQDC.Widget, PRect(@DstRect), Image.Handle, @SrcRect, QtAutoColor);
 finally
 Image.Free;
 end;
end;

{$endif}

{$ifdef LCLWin32}

procedure OpBitmapPaint32A(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer;
          SourcePrepared:boolean=false);
var
  Info: Windows.TBitmapInfo;
  hBmp: HBitmap;
  memDC: HDC;
  hOld: HGDIOBJ;
  BitsPtr: Pointer;
  DataSize: integer;
  Align: byte;
  bf: TBlendFunction = (BlendOp:AC_SRC_OVER;BlendFlags:0;
    SourceConstantAlpha:255;AlphaFormat:AC_SRC_ALPHA);
  TempBmp: TOPBitmap;
  tarr,sarr:PAPixel32;
  i:integer;
begin
 if (Bmp.Width=0) or (Bmp.Height=0) then exit;
  Bmp.PixelFormat := pf32bit;

  ZeroMemory(@Info, sizeof(Info));
  BitsPtr := nil;
  with Info.bmiHeader do
  begin
    biSize := SizeOf(BITMAPINFOHEADER);
    biWidth := Bmp.Width;
    biHeight := -Bmp.Height;
    biPlanes := 1;
    biBitCount := 32;
    biCompression := BI_RGB;
    biSizeImage := 0;
    biClrImportant := 0;
  end;

  hBmp := Windows.CreateDIBSection(Canvas.Handle, Info, DIB_RGB_COLORS,
    BitsPtr, 0, 0);

    DataSize:=((Info.bmiHeader.biWidth * info.bmiHeader.biBitCount + 31 ) shr 5 ) shl 2;
    DataSize:=DataSize * -info.bmiHeader.biHeight;

  If Not SourcePrepared Then
  begin
  tarr:=PAPixel32(BitsPtr);
  sarr:=TBitmapData32(Bmp.Data).RawArray;;
  for i := 0 to (DataSize div 4)-1 do
  begin
    tarr^[i].Blue := PreMultipliedTable[sarr^[i].Blue][sarr^[i].Alpha];
    tarr^[i].Red := PreMultipliedTable[sarr^[i].Red][sarr^[i].Alpha];
    tarr^[i].Green := PreMultipliedTable[sarr^[i].Green][sarr^[i].Alpha];
    tarr^[i].Alpha := sarr^[i].Alpha;
  end;
  end else Move(Bmp.ScanLine[0]^, BitsPtr^, DataSize);

  memDC := CreateCompatibleDC(Canvas.Handle);
  hOld := SelectObject(memDC, hBMp);

  AlphaBlend(Canvas.Handle,
    X, Y, Bmp.Width, Bmp.Height,
    memDC,
    0, 0, Bmp.Width, Bmp.Height,
    bf);

  SelectObject(MemDC, hOld);
  DeleteDC(memDC);
  DeleteObject(hBMp);
end;


procedure OpBitmapPaint32(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer);
var
  Info: Windows.TBitmapInfo;
begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
  Bmp.PixelFormat:=pf32bit;
  ZeroMemory(@Info, sizeof(Info));
  with Info.bmiHeader do
  begin
    biSize := SizeOf(BITMAPINFOHEADER);
    biWidth := Bmp.Width;
    biHeight := Bmp.Height;
    biPlanes := 1;
    biBitCount := 32;
    biCompression := BI_RGB;
    biSizeImage := 0;
    biClrImportant := 0;
  end;
 Windows.SetStretchBltMode(Canvas.Handle,COLORONCOLOR);
 Windows.StretchDIBits(Canvas.Handle, X, Pred(Y + Bmp.Height), Bmp.Width, -Bmp.Height,
   0, 0, Bmp.Width, Bmp.Height, Bmp.Scanline[0], Info, DIB_RGB_COLORS, SRCCOPY);
end;

{$endif}

{$ifdef LCLCarbon}


procedure OpBitmapPaint32(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer);
begin
  {
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
   Bmp.PixelFormat := pf32bit;
   CarbonBitmap := TCarbonBitmap.Create(Bitmap.Width, Bitmap.Height, 24, 32, cbaDWord, cbtRGB, Bmp.ScanLine[0], False);
   try
   CGImage := CarbonBitmap.CreateSubImage(Bounds(0, 0, Bitmap.Width, Bitmap.Height));
   TCarbonDeviceContext(Dest).DrawCGImage(X, Y, Bitmap.Width, Bitmap.Height, CGImage);
   CGImageRelease(CGImage);
   finally
   CarbonBitmap.Free;
   end;}
end;

Procedure OpBitmapPaint32A(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer;
          SourcePrepared:boolean=false);
 var
 CGImage: CGImageRef;
 CarbonBitmap: TCarbonBitmap;
begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
   Bmp.PixelFormat := pf32bit;
   CarbonBitmap := TCarbonBitmap.Create(Bitmap.Width, Bitmap.Height, 32, 32, cbaDWord, cbtARGB, Bmp.ScanLine[0], False);
   try
   CGImage := CarbonBitmap.CreateSubImage(Bounds(0, 0, Bitmap.Width, Bitmap.Height));
   TCarbonDeviceContext(Dest).DrawCGImage(X, Y, Bitmap.Width, Bitmap.Height, CGImage);
   CGImageRelease(CGImage);
   finally
   CarbonBitmap.Free;
   end;
end;
{$endif}

{$ifndef Op32_Supported}

procedure OpBitmapPaint32(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer);
begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
  LazBridge.AssignOpBitmapToCanvas(Bmp, Canvas, X, Y);
end;

Procedure OpBitmapPaint32A(Bmp: TOpBitmap; Canvas: Graphics.TCanvas; X, Y: integer;
          SourcePrepared:boolean=false);
begin
  if (Bmp.Width=0) or (Bmp.Height=0) then exit;
  LazBridge.AssignOpBitmapToCanvas(Bmp, Canvas, X, Y);
end;

{$endif}


end.

