unit opfppicture;

{$MODE objfpc}

interface

uses
  Classes, SysUtils, fpimage, opbitmap;

type

  { TOpFpPicture }

  TOpFpPicture = class(TFPCustomImage)
  private
    fOpBmp: TCanvasOPBitmap;
  protected
    procedure SetInternalColor(x, y: integer; const Value: TFPColor); override;
    function GetInternalColor(x, y: integer): TFPColor; override;
    procedure SetInternalPixel(x, y: integer; Value: integer); override;
    function GetInternalPixel(x, y: integer): integer; override;
  public
    constructor Create(AWidth, AHeight: integer); override;
    destructor Destroy; override;
    procedure SetSize(AWidth, AHeight: integer); override;
    procedure SyncSize; virtual;
    property OpBmp: TCanvasOPBitmap read fOpBmp;
  end;

procedure FpImageToOpBitmap(Fp:TFPCustomImage;Op:TOpBitmap);
procedure OpBitmapToFpImage(Op:TOpBitmap; Fp:TFPCustomImage);

implementation


{ TOpFpPicture }


procedure TOpFpPicture.SetInternalColor(x, y: integer; const Value: TFPColor);
var pix: Pixel32;
begin
  pix.Red := Value.red shr 8;
  pix.Green := Value.Green shr 8;
  pix.Blue := Value.blue shr 8;
  pix.Alpha := Value.alpha shr 8;
  TBitmapData32(fOpBmp.Data).RawArray^[Y * Width + X] := pix;
end;

function TOpFpPicture.GetInternalColor(x, y: integer): TFPColor;
var pix: Pixel32;
begin
  Pix := TBitmapData32(fOpBmp.Data).RawArray^[Y * Width + X];
  Result.red := Pix.Red + (Pix.Red shl 8);
  Result.green := Pix.Green + (Pix.Green shl 8);
  Result.blue := Pix.Blue + (Pix.Blue shl 8);
  Result.alpha := Pix.Alpha + (Pix.Alpha shl 8);
end;

procedure TOpFpPicture.SetInternalPixel(x, y: integer; Value: integer);
var pix: Pixel32;
  col: TFPColor;
begin
  col := Palette.Color[Value];
  pix.Red := col.red shr 8;
  pix.Green := col.Green shr 8;
  pix.Blue := col.blue shr 8;
  pix.Alpha := col.alpha shr 8;
  TBitmapData32(fOpBmp.Data).RawArray^[Y * Width + X] := pix;
end;

function TOpFpPicture.GetInternalPixel(x, y: integer): integer;
var pix: Pixel32;
  FpCol: TFPColor;
begin
  Pix := TBitmapData32(fOpBmp.Data).RawArray^[Y * Width + X];
  FpCol.red := Pix.Red + (Pix.Red shl 8);
  FpCol.green := Pix.Green + (Pix.Green shl 8);
  FpCol.blue := Pix.Blue + (Pix.Blue shl 8);
  FpCol.alpha := Pix.Alpha + (Pix.Alpha shl 8);
  Result := Palette.IndexOf(FpCol);
end;

constructor TOpFpPicture.Create(AWidth, AHeight: integer);
begin
  fOpBmp := TCanvasOPBitmap.Create;
  inherited Create(AWidth, AHeight);
end;

destructor TOpFpPicture.Destroy;
begin
  fOpBmp.free;
  inherited Destroy;
end;

procedure TOpFpPicture.SetSize(AWidth, AHeight: integer);
begin
  fOpBmp.PixelFormat := pf32bit;
  inherited SetSize(AWidth, AHeight);
  fOpBmp.Width := AWidth;
  fOpBmp.Height := AHeight;
end;

procedure TOpFpPicture.SyncSize;
begin
  inherited SetSize(OpBmp.Width, OpBmp.Height);
end;

procedure OpBitmapToFpImage(Op:TOpBitmap; Fp:TFPCustomImage);
var pix:pixel32;
col:TFPColor;
i,j:integer;
begin
Op.PixelFormat:=pf32bit;
fp.Width:=op.Width;
fp.Height:=op.Height;
for i:=0 to op.Height-1 do
 for j:=0 to op.Width-1 do
   begin
     Pix:=TBitmapData32(Op.Data).NativePixels[j,i];
     col.red:=Pix.Red+(Pix.Red shl 8);
     col.green:=Pix.Green+(Pix.Green shl 8);
     col.blue:=Pix.Blue+(Pix.Blue shl 8);
     col.alpha:=Pix.Alpha+(Pix.Alpha shl 8);
     fp.Colors[j,i]:=col;
   end;
end;

procedure FpImageToOpBitmap(Fp:TFPCustomImage;Op:TOpBitmap);
var pix:pixel32;
col:TFPColor;
i,j:integer;
begin
Op.Height:=0;
Op.PixelFormat:=pf32bit;
Op.Width:=fp.Width;
Op.Height:=fp.Height;
for i:=0 to fp.Height-1 do
 for j:=0 to fp.Width-1 do
   begin
     col:=fp.Colors[j,i];
     Pix.Red:=col.red shr 8;
     Pix.Green:=col.green shr 8;
     Pix.Blue:=col.blue shr 8;
     Pix.Alpha:=col.alpha shr 8;
     TBitmapData32(Op.Data).NativePixels[j,i]:=pix;
   end;
end;



end.
